/* 
 * smwidget.c --
 *
 *	This file implements viewport, an OpenGL/MESA widget.
 *
 * Portion of this code based on tkSquare.c
 *
 * Copyright (c) 1991-1994 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms.TclTk" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "copyright.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include "smInt.h"

/*
 * Prototypes for procedures defined externally
 */

XVisualInfo	*SmGetVisualInfo _ANSI_ARGS_ ((Display *dpy, Tk_Window tkwin, int viewport));

int             SmConfigureViewport _ANSI_ARGS_ ((Tcl_Interp *interp,
                                                  Tk_Canvas canvas, Tk_Item *itemPtr, int argc, char **argv, int flags));

void            SmRemovePortFromCell _ANSI_ARGS_ ((Cell *cell, PortItem *port));

void		SmCreateDrawable _ANSI_ARGS_ ((PortItem *port));

void		SmDisplayViewport _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, Display *dpy, Drawable dst,
						int x, int y, int width, int height));


/*
 * Prototypes for procedures defined in this file that are exported:
 */

void		SmViewportUpdate _ANSI_ARGS_((ClientData clientData));

/*
 * Prototypes for procedures defined in this file that are static:
 */

static void	ViewportCmdDeletedProc _ANSI_ARGS_((ClientData clientData));

static void	ViewportDestroy _ANSI_ARGS_((ClientData clientData));

static void	ViewportEventProc _ANSI_ARGS_((ClientData clientData, XEvent *eventPtr));

static int	ViewportWidgetCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
					       int argc, char **argv));
/*
 * Data structure declarations
 */

extern Tk_ConfigSpec widgetConfigSpecs[];



/*
 *--------------------------------------------------------------
 *
 * ViewportCmd --
 *
 * create a new viewport widget.
 *
 *--------------------------------------------------------------
 */

int
ViewportCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    int i, width, height;
    XGCValues gcValues;
    unsigned long mask;
    PortItem *port;
    Tk_Window tkwin;
    Tk_Window main = (Tk_Window) clientData;

    if (argc < 4) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " pathName width height ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[2], &width) != TCL_OK ||
	Tcl_GetInt(interp, argv[3], &height) != TCL_OK) {
	return TCL_ERROR;
    }

    if (width < 1 || height < 1) {
	Tcl_AppendResult(interp, "bad widget geometry: ", argv[2], "x", argv[3],
			 (char *) NULL);
	return TCL_ERROR;
    }

    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], (char *) NULL);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "Viewport");
    Tk_GeometryRequest(tkwin, width, height);

    /*
     * allocate and initialize the widget record.
     */

    port = (PortItem *) ckalloc(sizeof(PortItem));

    if (!(port->visinfo = SmGetVisualInfo(Tk_Display(tkwin), tkwin, 1))) {
	Tcl_AppendResult(interp, "failed to find a valid visual", (char *) NULL);
	Tk_DestroyWindow(tkwin);
	return TCL_ERROR;
    }

    port->tkwin = tkwin;
    port->dpy = Tk_Display(tkwin);
    port->interp = interp;
    port->widgetCmd = Tcl_CreateCommand(interp,
					Tk_PathName(port->tkwin), ViewportWidgetCmd,
					(ClientData) port, ViewportCmdDeletedProc);

    mask = GCFunction | GCPlaneMask;
    gcValues.function = GXcopy;
    gcValues.plane_mask = AllPlanes;

    port->redraw = 1;
    port->update = 0;
    port->canvas = NULL;
    port->initialized = 0;
    port->pm = None;
    port->pmGC = Tk_GetGC(tkwin, mask, &gcValues);
    port->glxpm = None;
    port->glxpmGC = None;
    port->outlineColor = NULL;
    port->fillColor = NULL;
    port->outlineGC = None;
    port->invalid = 0;
    port->cell = NULL;
    port->next = NULL;
    port->pick[0] = 0;
    port->ambientLight = NULL;
    port->ambient[0] = port->ambient[1] = port->ambient[2] = 0;
    port->ambient[3] = 1.0;
    port->depthcue.color = NULL;
    port->hsr = 1;
    port->epsilon = 1e-10;
    port->srcx = port->srcy = 0;
    port->srcw = port->srch = 0;
    port->w = 0;
    port->h = 0;
    port->invertedVisibleMask = 0;
    port->visibleMaskStr = NULL;

    for (i = 0; i < 8; i++) {
	port->lights[i].on = 0;
	port->lights[i].ambientColor = port->lights[i].diffuseColor = port->lights[i].specularColor = NULL;
    }
    Tcl_InitHashTable(&port->visibleHash, TCL_STRING_KEYS);

    Tk_CreateEventHandler(port->tkwin, ExposureMask|StructureNotifyMask,
			  ViewportEventProc, (ClientData) port);

    if (SmConfigureViewport(interp, (Tk_Canvas) tkwin, (Tk_Item *) port, argc-4, argv+4, 0) != TCL_OK) {
	Tk_DestroyWindow(port->tkwin);
	return TCL_ERROR;
    }

    interp->result = Tk_PathName(port->tkwin);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ViewportWidgetCmd --
 *
 * process the Tcl command that corresponds to a widget managed
 * by this module.
 *
 *--------------------------------------------------------------
 */

static int
ViewportWidgetCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about square widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    PortItem *port = (PortItem *) clientData;
    int result = TCL_OK;
    size_t length;
    char c;
    int width, height;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) port);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)
	&& (length >= 2)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " cget option\"",
			     (char *) NULL);
	    goto error;
	}
	result = Tk_ConfigureValue(interp, port->tkwin, widgetConfigSpecs, (char *) port, argv[2], 0);
    } else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)
	       && (length >= 2)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, port->tkwin, widgetConfigSpecs, (char *) port, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, port->tkwin, widgetConfigSpecs, (char *) port, argv[2], 0);
	} else {
	    result = SmConfigureViewport(interp, (Tk_Canvas) port->tkwin, (Tk_Item *) port, argc-2, argv+2,
					 TK_CONFIG_ARGV_ONLY);
	    Tk_SetInternalBorder(port->tkwin, port->width);
	}
    } else if ((c == 's') && (strncmp(argv[1], "size", length) == 0)
	       && (length >= 2)) {
	if (argc == 2) {
	    sprintf(interp->result, "%d %d", port->w, port->h);
	} else if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # size arguments: must be {width height}", (char *) NULL);
	    result = TCL_ERROR;
	} else if (Tcl_GetInt(interp, argv[2], &width) != TCL_OK ||
		   Tcl_GetInt(interp, argv[3], &height) != TCL_OK) {
	    result = TCL_ERROR;
	} else if (width < 1 || height < 1) {
	    Tcl_AppendResult(interp, "bad size value: width and height must be positive", (char *) NULL);
	    return TCL_ERROR;
	} else {
	    Tk_GeometryRequest(port->tkwin, width, height);
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\":  must be cget, configure, or size",
			 (char *) NULL);
	goto error;
    }
    if (!port->update) {
	Tk_DoWhenIdle(SmViewportUpdate, (ClientData) port);
	port->update = 1;
    }
    Tk_Release((ClientData) port);
    return result;

    error:
    Tk_Release((ClientData) port);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ViewportEventProc --
 *
 * process various events on viewports.
 *
 *--------------------------------------------------------------
 */

void
ViewportEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    PortItem *port = (PortItem *) clientData;
    Tk_Window tkwin = port->tkwin;

    if (eventPtr->type == Expose) {
	if (!port->update) {
	    Tk_DoWhenIdle(SmViewportUpdate, (ClientData) port);
	    port->update = 1;
	}
    } else if (eventPtr->type == ConfigureNotify) {
	if (!port->update) {
	    Tk_DoWhenIdle(SmViewportUpdate, (ClientData) port);
	    port->update = 1;
	}
	port->w = Tk_Width(tkwin);
	port->h = Tk_Height(tkwin);
	SmCreateDrawable(port);
    } else if (eventPtr->type == DestroyNotify) {
	if (port->tkwin != NULL) {
	    port->tkwin = NULL;
	    Tcl_DeleteCommand(port->interp,
			      Tcl_GetCommandName(port->interp,
						 port->widgetCmd));
	}
	if (port->update) {
	    Tk_CancelIdleCall(SmViewportUpdate, (ClientData) port);
	}
	Tk_EventuallyFree((ClientData) port, ViewportDestroy);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ViewportCmdDeletedProc --
 *
 * destroys a viewport widget.
 *
 *----------------------------------------------------------------------
 */

static void
ViewportCmdDeletedProc(clientData)
     ClientData clientData;	/* Pointer to widget record for widget. */
{
    PortItem *port = (PortItem *) clientData;
    Tk_Window tkwin = port->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
	port->tkwin = NULL;
	Tk_DestroyWindow(tkwin);
    }
}

/*
 *--------------------------------------------------------------
 *
 * SmViewportUpdate:
 *
 * redraw the contents of a viewport widget.
 *
 *--------------------------------------------------------------
 */

void
SmViewportUpdate(clientData)
     ClientData clientData;	/* Information about window. */
{
    PortItem *port = (PortItem *) clientData;
    Tk_Window tkwin = port->tkwin;

    port->update = 0;
    port->redraw = 1;
    if (!Tk_IsMapped(tkwin)) {
	return;
    }

#ifdef DEBUG_VIEW
    printf("display viewport widget: size = %dx%d (%dx%d)\n",
	   Tk_Width(tkwin), Tk_Height(tkwin), port->w, port->h);
#endif

    if ((port->w == 0)|| (port->h == 0)) return;

    SmDisplayViewport(NULL, (Tk_Item *) port, port->dpy, Tk_WindowId(port->tkwin),
		      0, 0, port->w, port->h);
}

/*
 *----------------------------------------------------------------------
 *
 * ViewportDestroy --
 *
 * clean up the internal structure of a viewport at a safe time
 * (when no-one is using it anymore).
 *
 *----------------------------------------------------------------------
 */

static void
ViewportDestroy(clientData)
     ClientData clientData;	/* Info about square widget. */
{
    int i;
    PortItem *port = (PortItem *) clientData;

    Tcl_DeleteHashTable(&port->visibleHash);

    Tk_FreeOptions(widgetConfigSpecs, (char *) port, port->dpy, 0);

    for (i = 0; i < 8; i++) {
	if (port->lights[i].on) {
	    Tk_FreeColor(port->lights[i].ambientColor);
	    Tk_FreeColor(port->lights[i].diffuseColor);
	    Tk_FreeColor(port->lights[i].specularColor);
	}
    }
    if (port->depthcue.color != NULL) {
	Tk_FreeColor(port->depthcue.color);
    }

    if (port->outlineGC != None) {
	Tk_FreeGC(port->dpy, port->outlineGC);
    }
    if (port->cell != NULL) {
	SmRemovePortFromCell(port->cell, port);
    }
    if (port->glxpm != None) {
	glXDestroyGLXPixmap(port->dpy, port->glxpm);
    }
    if (port->glxpmGC != None) {
	glXDestroyContext(port->dpy, port->glxpmGC);
    }
    if (port->pm != None) {
	XFreePixmap(port->dpy, port->pm);
    }
    if (port->pmGC != None) {
	Tk_FreeGC(port->dpy, port->pmGC);
    }
    ckfree((char *) port);
}

